import json
import re
import shlex
from asyncio import sleep as async_sleep
from io import BytesIO
from tempfile import SpooledTemporaryFile
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, Optional, Sequence, Union
from uuid import UUID, uuid4

import botocore.client
from clowmdb.models import WorkflowExecution, WorkflowMode
from fastapi import BackgroundTasks, HTTPException, UploadFile, status
from httpx import AsyncClient, ConnectError, ConnectTimeout
from mako.template import Template
from opentelemetry import trace
from PIL import Image, UnidentifiedImageError
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.config import settings
from app.crud import CRUDBucket, CRUDWorkflowExecution, CRUDWorkflowVersion
from app.git_repository.abstract_repository import GitRepository
from app.schemas.workflow_mode import WorkflowModeIn
from app.scm import SCM
from app.slurm.slurm_rest_client import SlurmClient

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import S3ServiceResource
else:
    S3ServiceResource = object

nextflow_command_template = Template(filename="mako_templates/nextflow_command.template")
# regex to find S3 files in parameters of workflow execution
s3_file_regex = re.compile(
    r"s3://(?!(((2(5[0-5]|[0-4]\d)|[01]?\d{1,2})\.){3}(2(5[0-5]|[0-4]\d)|[01]?\d{1,2})$))[a-z\d][a-z\d.-]{1,61}[a-z\d][^\"]*"
)

tracer = trace.get_tracer_provider().get_tracer(__name__)


def upload_icon(s3: S3ServiceResource, background_tasks: BackgroundTasks, icon: UploadFile) -> str:
    """
    Upload an icon to the icon bucket.

    Parameters
    ----------
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph.
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks.
    icon : fastapi.UploadFile
        Icon for the workflow version.

    Returns
    -------
    icon_slug : str
        Slug of the icon in the bucket
    """
    try:
        Image.open(icon.file)
    except UnidentifiedImageError:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="icon needs to be an image")
    icon_slug = f"{uuid4().hex}.png"
    # Write the icon to the icon bucket in a background task
    background_tasks.add_task(_process_and_upload_icon, s3=s3, icon_slug=icon_slug, icon_buffer=icon.file)
    return icon_slug


def _process_and_upload_icon(s3: S3ServiceResource, icon_slug: str, icon_buffer: BinaryIO) -> None:
    """
    Process the icon and upload it to the S3 Icon Bucket

    Parameters
    ----------
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph.
    icon_slug : str
        Slug of the icon
    icon_buffer : typing.BinaryIO
        Binary stream containing the image
    """
    im = Image.open(icon_buffer)
    im.thumbnail((64, 64))  # Crop to 64x64 image
    thumbnail_buffer = BytesIO()
    im.save(thumbnail_buffer, "PNG")  # save in buffer as PNG image
    thumbnail_buffer.seek(0)
    with tracer.start_as_current_span("s3_upload_workflow_version_icon") as span:
        span.set_attribute("icon", icon_slug)
        # Upload to bucket
        s3.Bucket(name=settings.ICON_BUCKET).Object(key=icon_slug).upload_fileobj(
            Fileobj=thumbnail_buffer, ExtraArgs={"ContentType": "image/png"}
        )


async def delete_remote_icon(s3: S3ServiceResource, db: AsyncSession, icon_slug: str) -> None:
    """
    Delete icon in S3 Bucket if there are no other workflow versions that depend on it

    Parameters
    ----------
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    icon_slug : str
        Name of the icon file.
    """
    # If there are no more Workflow versions that have this icon, delete it in the S3 ICON_BUCKET
    if not await CRUDWorkflowVersion.icon_exists(db, icon_slug):
        with tracer.start_as_current_span("s3_delete_workflow_version_icon") as span:
            span.set_attribute("icon", icon_slug)
            s3.Bucket(name=settings.ICON_BUCKET).Object(key=icon_slug).delete()


async def check_repo(
    repo: GitRepository, client: AsyncClient, modes: Optional[Sequence[Union[WorkflowModeIn, WorkflowMode]]] = None
) -> None:
    """
    Check if the necessary files are present in the git repository. If not, raises an HTTP Exception.

    Parameters
    ----------
    repo : app.git_repository.GitRepository
        Repo to check against
    client : httpx.AsyncClient
        Http client with an open connection.
    modes : List[app.schemas.workflow_mode.WorkflowModeIn | clowmdb.models.WorkflowMode] | None
        List of different mode of the workflow with different parameter schemas.
    """
    files = ["README.md", "CHANGELOG.md", "main.nf", "docs/usage.md", "docs/output.md"]
    if modes is not None and len(modes) > 0:
        # If there are modes, add their parameter schemas to the list
        for mode in modes:
            files.append(mode.schema_path)
    else:
        # Otherwise add default parameter schemas file location
        files.append("nextflow_schema.json")
    await repo.check_files_exist(
        files=files,
        client=client,
        raise_error=True,
    )


async def start_workflow_execution(
    s3: S3ServiceResource,
    db: AsyncSession,
    execution: WorkflowExecution,
    report_output_bucket: Optional[str],
    parameters: Dict[str, Any],
    git_repo: GitRepository,
    slurm_client: SlurmClient,
    scm_file_id: Optional[str] = None,
    workflow_entrypoint: Optional[str] = None,
) -> None:
    """
    Start a workflow on the Slurm cluster.

    Parameters
    ----------
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    execution : clowmdb.models.WorkflowExecution
        Workflow execution to execute.
    report_output_bucket : str | None
        Bucket where to save the Nextflow report.
    parameters : Dict[str, Any]
        Parameters for the workflow.
    git_repo : app.git_repository.abstract_repository.GitRepository
        Git repository of the workflow version.
    slurm_client : app.slurm.slurm_rest_client.SlurmClient
        Slurm Rest Client to communicate with Slurm cluster.
    scm_file_id : str | None
        ID of the SCM file for private git repositories.
    workflow_entrypoint : str | None
        Entrypoint for the workflow by specifying the `-entry` parameter
    """
    # Upload parameters to
    params_file_name = f"params-{execution.execution_id.hex}.json"
    with SpooledTemporaryFile(max_size=512000) as f:
        f.write(json.dumps(parameters).encode("utf-8"))
        f.seek(0)
        with tracer.start_as_current_span("s3_upload_workflow_execution_parameters") as span:
            span.set_attribute("workflow_execution_id", str(execution.execution_id))
            s3.Bucket(name=settings.PARAMS_BUCKET).Object(key=params_file_name).upload_fileobj(f)
    for key in parameters.keys():
        if isinstance(parameters[key], str):
            # Escape string parameters for bash shell
            parameters[key] = shlex.quote(parameters[key]).replace("$", "\$")

    # Check if the there is an SCM file for the workflow
    scm_file_name = None
    if scm_file_id is not None:
        scm_file_name = f"{scm_file_id}.scm"
        try:
            with tracer.start_as_current_span("s3_check_workflow_execution_parameters"):
                s3.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name).load()
            scm_file_id = f"repo{scm_file_id}"
        except botocore.client.ClientError:
            scm_file_id = None
            scm_file_name = None

    nextflow_command = nextflow_command_template.render(
        repo=git_repo,
        parameters=parameters,
        execution_id=execution.execution_id,
        configuration=settings.NX_CONFIG,
        nx_bin=settings.NX_BIN,
        scm_file_id=scm_file_id,
        report_output_bucket=report_output_bucket[5:]
        if report_output_bucket is not None and report_output_bucket.startswith("s3://")
        else report_output_bucket,
        workflow_entrypoint=workflow_entrypoint,
    )

    try:
        # Try to start the job on the slurm cluster
        slurm_job_id = await slurm_client.submit_job(
            nextflow_script=nextflow_command, execution_id=execution.execution_id, scm_file_name=scm_file_name
        )
        await CRUDWorkflowExecution.update_slurm_job_id(
            db, slurm_job_id=slurm_job_id, execution_id=execution.execution_id
        )
        await _monitor_proper_job_execution(
            db=db, slurm_client=slurm_client, execution_id=execution.execution_id, slurm_job_id=slurm_job_id
        )
    except (ConnectError, ConnectTimeout):  # pragma: no cover
        # Mark job as aborted when there is an error
        await CRUDWorkflowExecution.cancel(
            db, execution_id=execution.execution_id, status=WorkflowExecution.WorkflowExecutionStatus.ERROR
        )


async def _monitor_proper_job_execution(
    db: AsyncSession, slurm_client: SlurmClient, execution_id: UUID, slurm_job_id: int
) -> None:  # pragma: no cover
    """
    Checks every settings.SLURM_JOB_STATUS_CHECK_INTERVAL seconds if the slurm job is still running as long as
    the workflow execution in the database is not marked as finished.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    slurm_client : app.slurm.slurm_rest_client.SlurmClient
        Slurm Rest Client to communicate with Slurm cluster.
    execution_id : uuid.UUID
        ID of the workflow execution
    slurm_job_id : int
        ID of the slurm job to monitor
    """
    previous_span_link = None
    while True:
        await async_sleep(settings.SLURM_JOB_STATUS_CHECK_INTERVAL)
        with tracer.start_span("monitor_job", links=previous_span_link) as span:
            span.set_attributes({"execution_id": str(execution_id), "slurm_job_id": slurm_job_id})
            if await slurm_client.is_job_finished(slurm_job_id):
                execution = await CRUDWorkflowExecution.get(db, execution_id=execution_id)
                # Check if the execution is marked as finished in the database
                if execution is not None and execution.end_time is None:
                    # Mark job as finished with an error
                    await CRUDWorkflowExecution.cancel(
                        db, execution_id=execution_id, status=WorkflowExecution.WorkflowExecutionStatus.ERROR
                    )
                break
            previous_span_link = [trace.Link(span.get_span_context())]


async def check_active_workflow_execution_limit(db: AsyncSession, uid: str) -> None:
    """
    Check the number of active workflow executions of a usr and raise an HTTP exception if a new one would violate the
    limit of active workflow executions.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    uid : str
        ID of a user.
    """
    active_executions = await CRUDWorkflowExecution.list(
        db,
        uid=uid,
        status_list=WorkflowExecution.WorkflowExecutionStatus.active_workflows(),
    )
    if settings != -1 and len(active_executions) + 1 > settings.ACTIVE_WORKFLOW_EXECUTION_LIMIT:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f"The active workflow execution limit per user is {settings.ACTIVE_WORKFLOW_EXECUTION_LIMIT}",
        )


async def check_buckets_access(
    db: AsyncSession, parameters: Dict[str, Any], uid: str, report_bucket: Optional[str] = None
) -> None:
    """
    Check if the user has access to the buckets referenced in the workflow execution parameters.
    Raises an HTTP Exception if there is an issue.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    parameters : Dict[str, Any]
        Parameters of the workflow.
    uid : str
        UID of a user.
    report_bucket : str | None, default None
        Bucket or Path in bucket where the report should be written.
    """
    errors = []
    parameters_string = json.dumps(parameters)
    for match in s3_file_regex.finditer(parameters_string):
        path = match.group()[5:]
        error = await _check_bucket_access(db, uid, path)
        if error is not None:
            errors.append(error)
    # Check if the user has access to the bucket where the report should be written
    if report_bucket is not None:
        error = await _check_bucket_access(
            db, uid, report_bucket[5:] if report_bucket.startswith("s3://") else report_bucket
        )
        if error is not None:
            errors.append(error)
    if len(errors) > 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=f"Errors with bucket access: {','.join(errors)}"
        )


async def _check_bucket_access(db: AsyncSession, uid: str, bucket_path: str) -> Optional[str]:
    """
    Check if the bucket exists and the user has READWRITE access to it.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    uid : str
        UID of a user.
    bucket_path :  str
        Bucket plus optional file path, e.g. 'testbucket' or 'testbucket/folder/test.txt'

    Returns
    -------
    error : str | None
        If the check was successful, nothing will be returned, otherwise the error will be returned.
    """
    bucket = bucket_path.split("/")[0]
    file = "/".join(bucket_path.split("/")[1:])
    if not await CRUDBucket.check_bucket_exist(db, bucket):
        return f"Bucket {bucket} does not exist."
    if not await CRUDBucket.check_access(db, bucket, uid=uid, key=file if len(file) > 0 else None):
        error = f"The current user doesn't have READWRITE access to the bucket {bucket}"
        if len("file") > 0:
            error += f" and file/directory {file}"
        return error
    return None


def upload_scm_file(s3: S3ServiceResource, scm: SCM, scm_file_id: str) -> None:
    """
    Upload the SCM file of a private workflow into the PARAMS_BUCKET with the workflow id as key

    Parameters
    ----------
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph.
    scm : app.scm.SCM
        Python object representing the SCM file
    scm_file_id : str
        ID of the scm file for the name of the file in the bucket
    """
    with BytesIO() as handle:
        scm.serialize(handle)
        handle.seek(0)
        with tracer.start_as_current_span("s3_upload_workflow_credentials"):
            s3.Bucket(settings.PARAMS_BUCKET).Object(f"{scm_file_id}.scm").upload_fileobj(handle)
