from typing import List, Optional, Sequence, Union
from uuid import UUID

from clowmdb.models import WorkflowExecution
from opentelemetry import trace
from sqlalchemy import delete, func, or_, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload

from app.schemas.workflow_execution import DevWorkflowExecutionIn, WorkflowExecutionIn

tracer = trace.get_tracer_provider().get_tracer(__name__)


class CRUDWorkflowExecution:
    @staticmethod
    async def create(
        db: AsyncSession,
        execution: Union[WorkflowExecutionIn, DevWorkflowExecutionIn],
        owner_id: str,
        notes: Optional[str] = None,
    ) -> WorkflowExecution:
        """
        Create a workflow execution in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        execution : app.schemas.workflow_execution.WorkflowExecutionIn | DevWorkflowExecutionIn
            Workflow execution input parameters.
        owner_id : str
            User who started the workflow execution.
        notes : str | None, default None
            Notes to add to the workflow execution. Only usd if 'execution' has type 'DevWorkflowExecutionIn'.

        Returns
        -------
        workflow_execution : clowmdb.models.WorkflowExecution
            The newly created workflow execution
        """
        with tracer.start_as_current_span("db_create_workflow_execution") as span:
            if isinstance(execution, WorkflowExecutionIn):
                workflow_execution = WorkflowExecution(
                    user_id=owner_id,
                    workflow_version_id=execution.workflow_version_id,
                    notes=execution.notes,
                    slurm_job_id=-1,
                    _workflow_mode_id=execution.mode.bytes if execution.mode is not None else None,
                )
            else:
                workflow_execution = WorkflowExecution(
                    user_id=owner_id, workflow_version_id=None, notes=notes, slurm_job_id=-1
                )
            db.add(workflow_execution)
            await db.commit()
            await db.refresh(workflow_execution)
            span.set_attribute("workflow_execution_id", str(workflow_execution.execution_id))
            return workflow_execution

    @staticmethod
    async def get(db: AsyncSession, execution_id: Union[bytes, UUID]) -> Optional[WorkflowExecution]:
        """
        Get a workflow execution by its execution id from the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        execution_id : uuid.UUID | bytes
            ID of the workflow execution

        Returns
        -------
        workflow_execution : clowmdb.models.WorkflowExecution
            The workflow execution with the given id if it exists, None otherwise
        """
        with tracer.start_as_current_span("db_get_workflow_execution") as span:
            eid = execution_id.bytes if isinstance(execution_id, UUID) else execution_id
            stmt = (
                select(WorkflowExecution)
                .where(WorkflowExecution._execution_id == eid)
                .options(joinedload(WorkflowExecution.workflow_version))
            )
            span.set_attributes({"workflow_execution_id": str(execution_id), "sql_query": str(stmt)})
            return await db.scalar(stmt)

    @staticmethod
    async def list(
        db: AsyncSession,
        uid: Optional[str] = None,
        workflow_version_id: Optional[str] = None,
        status_list: Optional[List[WorkflowExecution.WorkflowExecutionStatus]] = None,
    ) -> Sequence[WorkflowExecution]:
        """
        List all workflow executions and apply filter.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        uid : str | None, default None
            Filter for the user who started the workflow execution.
        workflow_version_id : str | None, default None
            Filter for the workflow version
        status_list : List[clowmdb.models.WorkflowExecution.WorkflowExecutionStatus] | None, default None
            Filter for the status of the workflow executions.

        Returns
        -------
        workflow_executions : List[clowmdb.models.WorkflowExecution]
            List of all workflow executions with applied filters.
        """
        with tracer.start_as_current_span("db_list_workflow_executions") as span:
            stmt = select(WorkflowExecution).options(joinedload(WorkflowExecution.workflow_version))
            if uid is not None:
                span.set_attribute("uid", uid)
                stmt = stmt.where(WorkflowExecution.user_id == uid)
            if workflow_version_id is not None:
                span.set_attribute("git_commit_hash", workflow_version_id)
                stmt = stmt.where(WorkflowExecution.workflow_version_id == workflow_version_id)
            if status_list is not None:
                span.set_attribute("status", [stat.name for stat in status_list])
                stmt = stmt.where(or_(*[WorkflowExecution.status == status for status in status_list]))
            span.set_attribute("sql_query", str(stmt))
            executions = (await db.scalars(stmt)).all()
            return executions

    @staticmethod
    async def delete(db: AsyncSession, execution_id: Union[bytes, UUID]) -> None:
        """
        Delete a workflow execution from the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        execution_id : uuid.UUID | bytes
            ID of the workflow execution
        """
        with tracer.start_as_current_span("db_delete_workflow_execution") as span:
            eid = execution_id.bytes if isinstance(execution_id, UUID) else execution_id
            stmt = delete(WorkflowExecution).where(WorkflowExecution._execution_id == eid)
            span.set_attributes({"workflow_execution_id": str(execution_id), "sql_query": str(stmt)})
            await db.execute(stmt)
            await db.commit()

    @staticmethod
    async def cancel(
        db: AsyncSession,
        execution_id: Union[bytes, UUID],
        status: WorkflowExecution.WorkflowExecutionStatus = WorkflowExecution.WorkflowExecutionStatus.CANCELED,
    ) -> None:
        """
        Update the status of a workflow execution to CANCELED in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        execution_id : uuid.UUID | bytes
            ID of the workflow execution
        status : clowmdb.models.WorkflowExecution.WorkflowExecutionStatus, default WorkflowExecutionStatus.CANCELED
            Error status the workflow execution should get
        """
        with tracer.start_as_current_span("db_cancel_workflow_execution") as span:
            eid = execution_id.bytes if isinstance(execution_id, UUID) else execution_id
            stmt = (
                update(WorkflowExecution)
                .where(WorkflowExecution._execution_id == eid)
                .values(status=status.name, end_time=func.UNIX_TIMESTAMP())
            )
            span.set_attributes(
                {"workflow_execution_id": str(execution_id), "status": status.name, "sql_query": str(stmt)}
            )
            await db.execute(stmt)
            await db.commit()

    @staticmethod
    async def update_slurm_job_id(db: AsyncSession, execution_id: Union[bytes, UUID], slurm_job_id: int) -> None:
        """
        Update the status of a workflow execution to CANCELED in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        execution_id : uuid.UUID | bytes
            ID of the workflow execution
        slurm_job_id : int
            New slurm job ID
        """
        with tracer.start_as_current_span("db_update_workflow_execution_slurm_id") as span:
            eid = execution_id.bytes if isinstance(execution_id, UUID) else execution_id
            stmt = (
                update(WorkflowExecution)
                .where(WorkflowExecution._execution_id == eid)
                .values(slurm_job_id=slurm_job_id)
            )
            span.set_attributes(
                {"workflow_execution_id": str(execution_id), "slurm_job_id": slurm_job_id, "sql_query": str(stmt)}
            )
            await db.execute(stmt)
            await db.commit()
