from typing import List, Optional, Sequence, Union
from uuid import UUID

from clowmdb.models import WorkflowVersion, workflow_mode_association_table
from opentelemetry import trace
from sqlalchemy import desc, insert, or_, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload, selectinload

tracer = trace.get_tracer_provider().get_tracer(__name__)


class CRUDWorkflowVersion:
    @staticmethod
    async def get(
        db: AsyncSession,
        git_commit_hash: str,
        workflow_id: Optional[Union[bytes, UUID]] = None,
        populate_workflow: bool = False,
    ) -> Optional[WorkflowVersion]:
        """
        Get a workflow version by its commit git_commit_hash.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        git_commit_hash : str
            Git commit git_commit_hash of the version.
        workflow_id : UUID | bytes | None, default None
            Specify the workflow the version has to belong to.
        populate_workflow: boolean, default False
            Flag if to populate the workflow attribute.

        Returns
        -------
        user : clowmdb.models.WorkflowVersion | None
            The workflow version with the given git_commit_hash if it exists, None otherwise
        """
        with tracer.start_as_current_span("db_get_workflow_version") as span:
            span.set_attribute("git_commit_hash", git_commit_hash)

            stmt = (
                select(WorkflowVersion)
                .where(WorkflowVersion.git_commit_hash == git_commit_hash)
                .options(selectinload(WorkflowVersion.workflow_modes))
            )
            if populate_workflow:
                span.set_attribute("populate_workflow", True)
                stmt = stmt.options(joinedload(WorkflowVersion.workflow))
            if workflow_id is not None:
                span.set_attribute("workflow_id", str(workflow_id))
                wid = workflow_id if isinstance(workflow_id, bytes) else workflow_id.bytes
                stmt = stmt.where(WorkflowVersion._workflow_id == wid)
            span.set_attribute("sql_query", str(stmt))
            return await db.scalar(stmt)

    @staticmethod
    async def get_latest(
        db: AsyncSession, wid: Union[bytes, UUID], published: bool = True
    ) -> Optional[WorkflowVersion]:
        """
        Get the latest version of a workflow.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        wid : bytes | uuid.UUID
            Id of a workflow
        published : bool, default = True
            Get the latest versions that is published, otherwise get latest version overall

        Returns
        -------
        user : clowmdb.models.WorkflowVersion | None
            The latest workflow version of the given workflow if the workflow exists, None otherwise
        """
        with tracer.start_as_current_span("db_get_latest_workflow_version") as span:
            span.set_attribute("workflow_id", str(wid))
            stmt = (
                select(WorkflowVersion)
                .where(
                    WorkflowVersion._workflow_id == wid.bytes
                    if isinstance(wid, UUID)
                    else wid  # type: ignore[arg-type]
                )
                .order_by(desc(WorkflowVersion.created_at))
                .limit(1)
                .options(selectinload(WorkflowVersion.workflow_modes))
            )
            if published:
                span.set_attribute("only_published", True)
                stmt = stmt.where(
                    or_(
                        *[
                            WorkflowVersion.status == status
                            for status in [WorkflowVersion.Status.PUBLISHED, WorkflowVersion.Status.DEPRECATED]
                        ]
                    )
                )
            span.set_attribute("sql_query", str(stmt))
            return await db.scalar(stmt)

    @staticmethod
    async def list(
        db: AsyncSession, wid: Union[bytes, UUID], version_status: Optional[List[WorkflowVersion.Status]] = None
    ) -> Sequence[WorkflowVersion]:
        """
        List all versions of a workflow.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        wid : bytes | uuid.UUID
            Git commit git_commit_hash of the version.
        version_status : List[clowmdb.models.WorkflowVersion.Status] | None, default None
            Filter versions based on the status

        Returns
        -------
        user : List[clowmdb.models.WorkflowVersion]
            All workflow version of the given workflow
        """
        with tracer.start_as_current_span("db_list_workflow_versions") as span:
            span.set_attribute("workflow_id", str(wid))
            stmt = (
                select(WorkflowVersion)
                .options(selectinload(WorkflowVersion.workflow_modes))
                .where(
                    WorkflowVersion._workflow_id == wid.bytes
                    if isinstance(wid, UUID)
                    else wid  # type: ignore[arg-type]
                )
            )
            if version_status is not None and len(version_status) > 0:
                span.set_attribute("version_status", [stat.name for stat in version_status])
                stmt = stmt.where(or_(*[WorkflowVersion.status == status for status in version_status]))
            stmt = stmt.order_by(WorkflowVersion.created_at)
            span.set_attribute("sql_query", str(stmt))
            return (await db.scalars(stmt)).unique().all()

    @staticmethod
    async def create(
        db: AsyncSession,
        git_commit_hash: str,
        version: str,
        wid: Union[bytes, UUID],
        icon_slug: Optional[str] = None,
        previous_version: Optional[str] = None,
        modes: Optional[List[UUID]] = None,
    ) -> WorkflowVersion:
        """
        Create a new workflow version.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        git_commit_hash : str
            Git commit git_commit_hash of the version.
        version : str
            New version in string format
        wid : bytes | uuid.UUID
            ID of the corresponding workflow
        icon_slug : str | None, default None
            Slug of the icon
        previous_version : str | None, default None
            Git commit git_commit_hash of the previous version
        modes : List[UUID] | None, default None
            List of workflow mode IDs that are linked to the new workflow version

        Returns
        -------
        workflow_version : clowmdb.models.WorkflowVersion
            Newly create WorkflowVersion
        """
        with tracer.start_as_current_span("db_create_workflow_version") as span:
            span.set_attributes({"git_commit_version": git_commit_hash, "workflow_id": str(wid)})
            if modes is None:
                modes = []
            workflow_version = WorkflowVersion(
                git_commit_hash=git_commit_hash,
                version=version,
                _workflow_id=wid.bytes if isinstance(wid, UUID) else wid,
                icon_slug=icon_slug,
                previous_version_hash=previous_version,
            )
            db.add(workflow_version)
            if len(modes) > 0:
                span.set_attribute("mode_ids", [str(m) for m in modes])
                await db.commit()
                await db.execute(
                    insert(workflow_mode_association_table),
                    [
                        {"workflow_version_commit_hash": git_commit_hash, "workflow_mode_id": mode_id.bytes}
                        for mode_id in modes
                    ],
                )
            await db.commit()
            return workflow_version

    @staticmethod
    async def update_status(db: AsyncSession, git_commit_hash: str, status: WorkflowVersion.Status) -> None:
        """
        Update the status of a workflow version.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        git_commit_hash : str
            Git commit git_commit_hash of the version.
        status : clowmdb.models.WorkflowVersion.Status
            New status of the workflow version
        """
        with tracer.start_as_current_span("db_update_workflow_version_status") as span:
            span.set_attributes({"status": status.name, "git_commit_version": git_commit_hash})
            stmt = (
                update(WorkflowVersion)
                .where(WorkflowVersion.git_commit_hash == git_commit_hash)
                .values(status=status.name)
            )
            span.set_attribute("sql_query", str(stmt))
            await db.execute(stmt)
            await db.commit()

    @staticmethod
    async def update_icon(db: AsyncSession, git_commit_hash: str, icon_slug: Optional[str] = None) -> None:
        """
        Update the icon slug for a workflow version.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        git_commit_hash : str
            Git commit git_commit_hash of the version.
        icon_slug : str | None, default None
            The new icon slug
        """
        with tracer.start_as_current_span("db_update_workflow_version_icon") as span:
            stmt = (
                update(WorkflowVersion)
                .where(WorkflowVersion.git_commit_hash == git_commit_hash)
                .values(icon_slug=icon_slug)
            )
            span.set_attributes(
                {
                    "git_commit_hash": git_commit_hash,
                    "icon_slug": icon_slug if icon_slug else "None",
                    "sql_query": str(stmt),
                }
            )
            await db.execute(stmt)
            await db.commit()

    @staticmethod
    async def icon_exists(db: AsyncSession, icon_slug: str) -> bool:
        """
        Check if there is a workflow version that depends on a icon.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        icon_slug : str
            The icon slug to search for

        Returns
        -------
        exists : bool
            Flag if a version exists that depends on the icon
        """
        with tracer.start_as_current_span("db_check_workflow_version_icon_exists") as span:
            stmt = select(WorkflowVersion).where(WorkflowVersion.icon_slug == icon_slug).limit(1)
            span.set_attributes({"icon_slug": icon_slug, "sql_query": str(stmt)})
            version_with_icon = await db.scalar(stmt)
            return version_with_icon is not None
