from functools import cached_property
from typing import Dict, Generator, Optional
from urllib.parse import quote, urlparse

from httpx import AsyncClient, Auth, Request, Response
from pydantic import AnyHttpUrl

from .abstract_repository import GitRepository


class BearerAuth(Auth):
    def __init__(self, token: str):
        self.token = token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request


class GitlabRepository(GitRepository):
    """
    Implementation for a Gitlab Repository
    """

    @property
    def token(self) -> Optional[str]:
        return self._token

    @property
    def provider(self) -> str:
        return "gitlab"

    @property
    def repo_url(self) -> str:
        return self.url

    @property
    def git_commit_hash(self) -> str:
        return self.commit

    @cached_property
    def request_auth(self) -> Optional[BearerAuth]:
        if self._token is not None:
            return BearerAuth(token=self._token)
        return None

    @cached_property
    def request_headers(self) -> Dict[str, str]:
        return {}

    def __init__(self, url: str, git_commit_hash: str, token: Optional[str] = None):
        self.url = url
        parse_result = urlparse(url)
        self.domain = parse_result.netloc
        self.project = parse_result.path[1:]
        self.commit = git_commit_hash
        self._token = token

    def check_file_url(self, filepath: str) -> AnyHttpUrl:
        return AnyHttpUrl.build(
            scheme="https",
            host=self.domain,
            path="/".join(
                ["api", "v4", "projects", quote(self.project, safe=""), "repository", "files", quote(filepath, safe="")]
            ),
            query=f"ref={quote(self.commit)}",
        )

    async def download_file_url(self, filepath: str, client: Optional[AsyncClient] = None) -> AnyHttpUrl:
        return AnyHttpUrl.build(
            scheme="https",
            host=self.domain,
            path="/".join(
                [
                    "api",
                    "v4",
                    "projects",
                    quote(self.project, safe=""),
                    "repository",
                    "files",
                    quote(filepath, safe=""),
                    "raw",
                ]
            ),
            query=f"ref={quote(self.commit)}",
        )

    def __repr__(self) -> str:
        url = AnyHttpUrl.build(scheme="https", host=self.domain, path=self.project)
        return f"Gitlab(repo={url} commit={self.commit}))"
