from os import environ
from typing import Optional
from uuid import UUID

import dotenv
from fastapi import status
from httpx import AsyncClient
from opentelemetry import trace

from app.core.config import settings

tracer = trace.get_tracer_provider().get_tracer(__name__)

dotenv.load_dotenv()
base_env = {key: val for key, val in environ.items() if key.startswith("NXF_")}


class SlurmClient:
    def __init__(self, client: AsyncClient, version: str = "v0.0.38"):
        """
        Initialize the client to communicate with a Slurm cluster.

        Parameters
        ----------
        client : httpx.AsyncClient
            Async HTTP client with an open connection.
        version : str, default "v0.0.38"
            Version of the Slurm Rest API
        """
        self._client = client
        self._headers = {"X-SLURM-USER-TOKEN": settings.SLURM_TOKEN, "X-SLURM-USER-NAME": settings.SLURM_USER}
        self.version = version

    async def submit_job(self, nextflow_script: str, execution_id: UUID, scm_file_name: Optional[str] = None) -> int:
        """
        Submit a job to the slurm cluster.

        Parameters
        ----------
        nextflow_script : str
            Script to execute on the slurm cluster.
        execution_id : uuid.UUID
            ID of the workflow execution.
        scm_file_name : str | None, default None
            Name of a SCM file if the workflow is in a private repository

        Returns
        -------
        slurm_job_id : int
            Slurm job ID of submitted job.
        """
        env = base_env.copy()
        env["TOWER_WORKSPACE_ID"] = execution_id.hex[:16]
        # Reference a SCM file for a private repository
        if scm_file_name is not None:
            env["NXF_SCM_FILE"] = f"{settings.PARAMS_BUCKET_MOUNT_PATH}/{scm_file_name}"

        body = {
            "script": nextflow_script,
            "job": {
                "name": str(execution_id),
                "current_working_directory": settings.SLURM_WORKING_DIRECTORY,
                "requeue": False,
                "environment": env,
            },
        }

        with tracer.start_as_current_span("slurm_submit_job"):
            response = await self._client.post(
                f"{settings.SLURM_ENDPOINT}slurm/{self.version}/job/submit", headers=self._headers, json=body
            )
        return int(response.json()["job_id"])

    async def cancel_job(self, job_id: int) -> None:
        """
        Cancel a Slurm job on the cluster.

        Parameters
        ----------
        job_id : int
            ID of the job to cancel.
        """
        with tracer.start_as_current_span("slurm_cancel_job"):
            await self._client.delete(
                f"{settings.SLURM_ENDPOINT}slurm/{self.version}/job/{job_id}", headers=self._headers
            )

    async def is_job_finished(self, job_id: int) -> bool:
        """
        Check if the job with the given is completed

        Parameters
        ----------
        job_id : int
            ID of the job to cancel.

        Returns
        -------
        finished : bool
            Flag if the job is finished
        """
        with tracer.start_as_current_span("slurm_check_job_status") as span:
            response = await self._client.get(
                f"{settings.SLURM_ENDPOINT}slurm/{self.version}/job/{job_id}", headers=self._headers
            )
            span.set_attribute("slurm.job-status.request.code", response.status_code)
            if response.status_code != status.HTTP_200_OK:
                return True
            try:  # pragma: no cover
                job_state = response.json()["jobs"][0]["job_state"]
                span.set_attribute("slurm.job-status.state", job_state)
                return job_state == "COMPLETED" or job_state == "FAILED" or job_state == "CANCELLED"
            except (KeyError, IndexError) as ex:
                span.record_exception(ex)
                return True
